import numpy as np
import os
os.environ['TRANSFORMERS_CACHE'] = 'cache'  # Must be before transformers import.
import pandas as pd
import sklearn
from sklearn import cluster, linear_model, metrics, multiclass
from sklearn.model_selection import train_test_split
import torch
import transformers

from . import utils
from .task import Task


class ClassificationTask(Task):
    def __init__(self, train, val, test):
        train = train if train is not None else pd.DataFrame([], columns=['text', 'label'])
        val = val if val is not None else pd.DataFrame([], columns=['text', 'label'])
        (
            self.train_embeds,
            self.train_labels,
            self.val_embeds,
            self.val_labels,
            self.test_embeds
        ) = self._load_data(train, val, test)

        self.classes = list(set(self.train_labels + self.val_labels))
        self.train_labels = sklearn.preprocessing.label_binarize(self.train_labels, classes=self.classes).tolist()
        if len(self.train_labels[0]) == 1:  # Quirk of label_binarize: binary cases must be expanded from a single column vector.
            self.train_labels = [[1, 0] if x == [0] else [0, 1] for x in self.train_labels]
        if len(self.val_labels) > 0:
            self.val_labels = sklearn.preprocessing.label_binarize(self.val_labels, classes=self.classes).tolist()
            if len(self.val_labels[0]) == 1:  # Quirk of label_binarize: binary cases must be expanded from a single column vector.
                self.val_labels = [[1, 0] if x == [0] else [0, 1] for x in self.val_labels]

        self.model = self._load_model()
    
    def _load_data(self, train, val, test):
        text = train.text.tolist() + val.text.tolist() + test.text.tolist()  # Must embed togeter to match sequence length.

        # Tokenize.
        tokenizer = transformers.BertTokenizer.from_pretrained('bert-base-uncased')
        token_ids = tokenizer(text, return_tensors='pt', padding=True)
        
        # Embed.
        embed_model = transformers.BertModel.from_pretrained('bert-base-uncased')
        embed_model.eval()
        with torch.no_grad():
            embeds = embed_model(**token_ids)[0]
        embeds = embeds.mean(axis=1)

        return embeds[:len(train)], train.label.tolist(), embeds[len(train):len(train) + len(val)], val.label.tolist(), embeds[len(train) + len(val):]

    def _load_model(self):
        return multiclass.OneVsRestClassifier(linear_model.LogisticRegression(random_state=0, max_iter=1000))
    
    def train(self):
        self.model.fit(self.train_embeds, self.train_labels)

    def evaluate(self):
        if len(self.val_embeds) == 0:
            return None

        val_preds = self.model.predict(self.val_embeds)
        val_probs = self.model.predict_proba(self.val_embeds)

        accs = {}
        precs = {}
        recs = {}
        f1s = {}
        aurocs = {}
        auprs = {}

        # Get one-vs-all classification metrics for each class.
        for i, class_name in enumerate(self.classes):
            class_bin = [0 for _ in range(len(self.classes))]
            class_bin[i] = 1

            labels_bin = [1 if j == class_bin else 0 for j in self.val_labels]
            preds_bin = [1 if j == class_bin else 0 for j in val_preds.tolist()]

            accs[class_name] = metrics.accuracy_score(labels_bin, preds_bin).item()
            precs[class_name] = metrics.precision_score(labels_bin, preds_bin, zero_division=0).item()
            recs[class_name] = metrics.recall_score(labels_bin, preds_bin, zero_division=0).item()
            f1s[class_name] = metrics.f1_score(labels_bin, preds_bin, zero_division=0).item()

            fpr, tpr, _ = metrics.roc_curve(labels_bin, val_probs[:,i])
            aurocs[class_name] = metrics.auc(fpr, tpr).item()

            precision, recall, _ = metrics.precision_recall_curve(labels_bin, val_probs[:,i])
            auprs[class_name] = metrics.auc(recall, precision).item()

        return {
            'acc': accs,
            'prec': precs,
            'rec': recs,
            'f1': f1s,
            'auroc': aurocs,
            'aupr': auprs,
        }
    
    def predict(self):
        predictions = self.model.predict(self.test_embeds)
        out_predictions = []
        for i in range(len(predictions)):
            class_i = np.argmax(predictions[i])
            out_predictions.append(self.classes[class_i])
        return out_predictions


class KMeansClassificationTask(ClassificationTask):
    def __init__(self, test):
        (_, _, _, _, self.test_embeds) = self._load_data(
            pd.DataFrame([], columns=['text', 'label']),
            pd.DataFrame([], columns=['text', 'label']),
            test
        )
        self.model = self._load_model()
    
    def _load_data(self, train, val, test):
        return super()._load_data(train, val, test)

    def _load_model(self):
        return cluster.KMeans(n_clusters=2, random_state=0, max_iter=1000)
    
    def train(self):
        self.model.fit(self.test_embeds)
    
    def evaluate(self):
        # KMeans task is default task when no train/val data is provided.
        # Thus, it should not have an evaluate function.
        pass
    
    def predict(self):
        return self.model.predict(self.test_embeds)


def classify(text_to_label, prelabeled_text=None, output_metrics=False):
    assert isinstance(text_to_label, pd.DataFrame), 'Make sure you are passing in pandas DataFrames.'
    assert 'text' in text_to_label.columns, 'Make sure the text column in your pandas DataFrame is named "text".'

    if prelabeled_text is not None:
        assert isinstance(prelabeled_text, pd.DataFrame), 'Make sure you are passing in pandas DataFrames.'
        assert 'text' in prelabeled_text.columns and 'label' in prelabeled_text.columns, \
            'Make sure the text column in your pandas DataFrame is named "text" and the label column is named "label"'

    if output_metrics:
        assert prelabeled_text is not None, 'In order to output an estimate of performance with output_metrics, prelabeled_text must be provided.'

    # Perform KMeans if no training data is given.
    if prelabeled_text is None:
        task = KMeansClassificationTask(text_to_label)
        task.train()
        pred_results = task.predict()
        results = text_to_label.copy()
        results['label_from_tailwiz'] = pred_results
        return results

    if len(prelabeled_text) < 3:
        raise ValueError('prelabeled_text has too few examples. At least 3 are required.')

    num_unique_classes = len(prelabeled_text.label.unique())
    if num_unique_classes <= 1:
        raise ValueError('prelabeled_text contains examples from just one class. Examples from at least 2 classes are required.')

    # Try 10 times to get a proper split. Sometimes, a split will cause all training examples to be
    # in the same class, which will error.
    classify_task_out = None
    split_attempt = 0
    while split_attempt < 10 and classify_task_out is None:
        train, val = sklearn.model_selection.train_test_split(prelabeled_text, test_size=0.2)
        num_unique_classes_in_train = len(train.label.unique())
        if num_unique_classes_in_train < 2:
            split_attempt += 1
            continue
        classify_task_out = ClassificationTask(train, val, text_to_label)
        classify_task_out.train()
    
    if classify_task_out is None:
        raise ValueError('''The provided prelabeled_text examples were not diverse enough to estimate performance.
        Try balancing your prelabeled_text examples by adding more examples of each class.''')
    
    pred_results = classify_task_out.predict()
    results = text_to_label.copy()
    results['label_from_tailwiz'] = pred_results

    if output_metrics:
        metrics_out = classify_task_out.evaluate()

    return (results, metrics_out) if output_metrics else results



import os
os.environ['TRANSFORMERS_CACHE'] = 'cache'  # Must be before transformers import.
import pandas as pd
import torch
import transformers
import evaluate
from sklearn.model_selection import train_test_split
import transformers

from .task import Task

class ParsingTask(Task):
    def __init__(self, train, val, test):
        self.tokenizer = transformers.BertTokenizer.from_pretrained('bert-large-uncased-whole-word-masking-finetuned-squad')
        (
            self.train_dataset,
            self.val_dataset,
            self.test_dataset
        ) = self._load_data(train, val, test)
        self.model = self._load_model()

    def _load_data(self, train, val, test):
        # Tokenize.
        train_tokens = None if train is None else self.tokenizer(train.prompt.tolist(), train.context.tolist(), train.label.tolist(), padding=True, return_tensors='pt')
        
        val_tokens = None if val is None else self.tokenizer(val.prompt.tolist(), val.context.tolist(), val.label.tolist(), padding=True, return_tensors='pt')

        test_tokens = self.tokenizer(test.prompt.tolist(), test.context.tolist(), padding=True, return_tensors='pt')

        # Helper function that finds the start and end positions of the label within the context.
        # By default sets the start and end positions to the first occurrence of the label.
        def get_target_start_end_pos(tokens, contexts, labels):
            target_start_pos = []
            target_end_pos = []
            for i, (text_tokenized, label_tokenized, context, label) in enumerate(zip(tokens['input_ids'], tokens['labels'], contexts, labels)):
                if label_tokenized[0] == 101 and label_tokenized[1] == 102:
                    target_start_pos.append(0)
                    target_end_pos.append(0)
                else:
                    label_tokenized_nopad = label_tokenized[label_tokenized.nonzero()].squeeze()
                    label_tokenized_nospectoks = label_tokenized_nopad[1:-1]

                    for j in range(len(text_tokenized) - len(label_tokenized_nospectoks)):  # Search from start for label
                        if (text_tokenized[j:j+len(label_tokenized_nospectoks)] == label_tokenized_nospectoks).all():
                            target_start_pos.append(j)
                            target_end_pos.append(j+len(label_tokenized_nospectoks))
                            break
                if len(target_start_pos) != i + 1:
                    raise ValueError(f'''We could not find the label:
                        
                        {label}

in the context:
                        
                        {context}

Either: (A) only include labels where the label is extracted exactly from the context as whole words, or (B) use the generate()
function, which does not require the labels do be found in the context.''')
            target_start_pos = torch.tensor(target_start_pos)
            target_end_pos = torch.tensor(target_end_pos)
            return target_start_pos, target_end_pos

        if train_tokens is not None:
            train_target_start_pos, train_target_end_pos = get_target_start_end_pos(train_tokens, train.context.tolist(), train.label.tolist())
            train_tokens['start_positions'] = train_target_start_pos
            train_tokens['end_positions'] =  train_target_end_pos
            train_tokens = {k:v for k, v in train_tokens.items() if k != 'labels'}

        if val_tokens is not None:
            val_target_start_pos, val_target_end_pos = get_target_start_end_pos(val_tokens, val.context.tolist(), val.label.tolist())
            val_tokens['start_positions'] = val_target_start_pos
            val_tokens['end_positions'] =  val_target_end_pos
            val_tokens = {k:v for k, v in val_tokens.items() if k != 'labels'}

        return train_tokens, val_tokens, test_tokens

    def _load_model(self):
        return transformers.BertForQuestionAnswering.from_pretrained('bert-large-uncased-whole-word-masking-finetuned-squad')

    def _decode_pos2strs(self, start_pos, end_pos, input_ids):
        strs = []
        for i, (j, k) in enumerate(zip(start_pos, end_pos)):
            pred_tokens = input_ids[i, j:k+1]
            strs.append(self.tokenizer.decode(pred_tokens, skip_special_tokens=True))
        return strs

    def train(self):
        class BLUWWMFSDataset(torch.utils.data.Dataset):
            def __init__(self, data):
                self.data = data
            
            def __len__(self):
                return len(self.data['input_ids'])
            
            def __getitem__(self, index):
                return {k: self.data[k][index] for k in self.data.keys()}

        args = transformers.TrainingArguments(
            'cache/bert-qa',
            num_train_epochs=10,
            evaluation_strategy='epoch',
            save_strategy='epoch',
            load_best_model_at_end=True,
            metric_for_best_model='eval_loss',
            logging_steps=1,
        )
        trainer = transformers.Trainer(
            model=self.model,
            args=args,
            train_dataset=BLUWWMFSDataset(self.train_dataset),
            eval_dataset=BLUWWMFSDataset(self.val_dataset)
        )
        trainer.train()

    def evaluate(self):
        # Remove 'labels' key from validation dataset for prediction.
        outputs = self.model.to(self.val_dataset['input_ids'].device)(**self.val_dataset)

        predictions = self._decode_pos2strs(outputs['start_logits'].argmax(1), outputs['end_logits'].argmax(1), self.val_dataset['input_ids'])
        references = self._decode_pos2strs(self.val_dataset['start_positions'], self.val_dataset['end_positions'], self.val_dataset['input_ids'])

        metrics_funcs = [
            evaluate.load('exact_match'),
            evaluate.load('rouge')
        ]
        metrics = {}
        for metric in metrics_funcs:
            metrics.update(metric.compute(predictions=predictions, references=references))
        for k in metrics.keys():
            metrics[k] = metrics[k].item()
        return metrics

    def predict(self):
        outputs = self.model.to(self.test_dataset['input_ids'].device)(**self.test_dataset)
        return self._decode_pos2strs(outputs['start_logits'].argmax(1), outputs['end_logits'].argmax(1), self.test_dataset['input_ids'])


def parse(text_to_label, prelabeled_text=None, output_metrics=False):
    assert isinstance(text_to_label, pd.DataFrame), 'Make sure you are passing in pandas DataFrames.'
    assert 'prompt' in text_to_label.columns and 'context' in text_to_label.columns, \
        'Make sure the prompt column in your pandas DataFrame is named "prompt" and the context column is named "context".'

    if prelabeled_text is not None:
        assert isinstance(prelabeled_text, pd.DataFrame), 'Make sure you are passing in pandas DataFrames.'
        assert 'prompt' in prelabeled_text.columns and 'context' in prelabeled_text.columns and 'label' in prelabeled_text.columns, \
            'Make sure the prompt column in your pandas DataFrame is named "prompt", the context column is named "context", and the label column is named "label".'
    if output_metrics:
        assert prelabeled_text is not None, 'In order to output an estimate of performance with output_metrics, prelabeled_text must be provided.'

    if prelabeled_text is None:
        parse_task_out = ParsingTask(None, None, text_to_label)
        pred_results = parse_task_out.predict()
    else:
        assert len(prelabeled_text) >= 2, 'At least 2 rows of prelabeled data must be given.'
        train, val = train_test_split(prelabeled_text, test_size=0.2)
        parse_task_out = ParsingTask(train, val, text_to_label)
        parse_task_out.train()
        pred_results = parse_task_out.predict()
    
    results = text_to_label.copy()
    results['label_from_tailwiz'] = pred_results

    if output_metrics:
        metrics = parse_task_out.evaluate()

    return (results, metrics) if output_metrics else results
